package test;

import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;
import java.util.PriorityQueue;

import javax.imageio.ImageIO;

public class Renderer {

	public static final double SPHERE_RADIUS = 1.0;
	public static final Vector LIGHT_POS = new Vector(2, 1.5, 0);

	public static final Vector NORMAL_LEFT = new Vector(-1, 0, 0);
	public static final Vector NORMAL_RIGHT = new Vector(1, 0, 0);
	public static final Vector NORMAL_UP = new Vector(0, 1, 0);
	public static final Vector NORMAL_DOWN = new Vector(0, -1, 0);
	public static final Vector NORMAL_FRONT = new Vector(0, 0, -1);
	public static final Vector NORMAL_BACK = new Vector(0, 0, 1);

	public static final Color RED = new Color(1, 0, 0);
	public static final Color BLUE = new Color(0, 0, 0.1);
	public static final Color BROWN = new Color(0.6, 0.3, 0.15);
	public static final Color WHITE = new Color(1, 1, 1);
	public static final Color BLACK = new Color(0, 0, 0);
	public static final Color GRAY = new Color(0.5, 0.5, 0.5);

	static Vector spherePos = new Vector(0, 0, -3);

	private static Color[][] PARQUET_TEXTURE = loadTexture("texture1.jpg"),
			CEILING_TEXTURE = loadTexture("texture2.jpg");

	public static class Ray {
		Vector pos, dir;
	}

	// loads a texture from a picture
	private static Color[][] loadTexture(String fileName) {
		Color[][] texture = null;
		try {
			URL url = Renderer.class.getResource("/" + fileName);
			BufferedImage img = ImageIO.read(url);
			int w = img.getWidth(), h = img.getHeight();
			int[] buffer = new int[w * h];
			img.getRGB(0, 0, w, h, buffer, 0, w);
			texture = new Color[w][h];
			int k = 0;
			for (int i = 0; i < w; i++)
				for (int j = 0; j < h; j++)
					texture[i][j] = new Color(buffer[k++]);
		} catch (IOException e) {
			e.printStackTrace();
		}
		return texture;
	}

	/*
	 * Determines whether the given ray hits the sphere. If yes, adds the hit
	 * information to the queue.
	 */
	private static void checkForSphereHit(Ray r, PriorityQueue<Hit> hits) {
		Vector d = spherePos.sub(r.pos);

		/*
		 * Wikipedia says: d = (l*c) +- sqrt((l*c)² - c² + r²) where l is the
		 * ray direction (ray.dir), c is the sphere's centre (spherePos) and r
		 * is the radius. This is for a ray originating at (0,0,0), so we
		 * subtract r.pos from spherePos first.
		 */
		double tmp = r.dir.dotProduct(d); // l*c
		double det = tmp * tmp - d.getLengthSquared() + SPHERE_RADIUS
				* SPHERE_RADIUS;

		if (det >= 0) {// sphere hit!
			double sqrtDet = Math.sqrt(det);
			double dist = tmp - sqrtDet;
			if (dist < 0)
				dist = tmp + sqrtDet;
			if (dist < 0)
				return;
			Hit hit = new Hit();
			hit.dist = dist;
			hit.pos = r.pos.add(r.dir.scale(dist));
			hit.normal = hit.pos.sub(spherePos).normalize();
			hit.color = RED;
			hit.shininess = 32;
			hit.reflectivity = 0.5;
			hits.offer(hit);
		}
	}

	/*
	 * Determines whether the given ray hits the sphere. If yes, returns the
	 * distance to the ray's origin, otherwise NaN.
	 */
	private static double checkForSphereHitDistOnly(Ray r) {
		Vector d = spherePos.sub(r.pos);

		double tmp = r.dir.dotProduct(d); // l*c
		double det = tmp * tmp - d.getLengthSquared() + SPHERE_RADIUS
				* SPHERE_RADIUS;
		double sqrtDet = Math.sqrt(det);
		double dist = tmp - sqrtDet;
		if (dist < 0)
			dist = tmp + sqrtDet;
		if (dist < 0)
			dist = Double.POSITIVE_INFINITY;
		return dist;
	}

	/*
	 * Checks whether the given ray hits the wall described by the given plane
	 * (through wallNormal and distance of the origin to the wall. Texture and
	 * colour can be set. If the wall is hit, the corresponding hit information
	 * is added to the queue.
	 */
	private static void checkForWallHit(Ray r, Vector wallNormal,
			double planeOriginDist, Color color, Color[][] texture,
			PriorityQueue<Hit> hits) {

		/*
		 * the cosine of the angle between the viewing direction and the surface
		 * normal of the wall.
		 */
		double cosVN = r.dir.dotProduct(wallNormal);
		if (cosVN >= 0)
			return;

		// the distance of the ray's origin to the wall
		double projDist = r.pos.dotProduct(wallNormal) + planeOriginDist;
		// the distance from the origin at which the ray hits the wall
		double dist = -projDist / cosVN;
		if (dist < 0)
			return;

		Hit hit = new Hit();
		hit.dist = dist;
		hit.pos = r.pos.add(r.dir.scale(dist));
		hit.normal = wallNormal;
		hit.color = color;
		hit.texture = texture;
		hits.offer(hit);
	}

	/*
	 * Determines whether the light is occluded by the sphere from the position
	 * of the given hit, i.e. whether the given hit is not in shade.
	 */
	private static boolean isOccluded(Hit hit) {
		/*
		 * launch a shadow ray from the light source to the position of the hit.
		 */
		Ray shadowRay = new Ray();
		Vector lightVector = hit.pos.sub(LIGHT_POS);
		shadowRay.pos = LIGHT_POS;
		shadowRay.dir = lightVector.normalize();
		/*
		 * Determine whether we hit the sphere and if yes, if we hit it before
		 * we hit the target object.
		 */
		double dist = checkForSphereHitDistOnly(shadowRay);
		return dist * dist < lightVector.getLengthSquared() - 0.001;
	}

	private static Color doLighting(Hit hit, Ray r) {
		// the direction at which the light hits the surface
		Vector lightDir = hit.pos.sub(LIGHT_POS).normalize();

		Color ambientLight = hit.color.scale(0.15);
		double diffuse = 0, specular = 0;

		// perform diffuse and specular lighting
		if (!isOccluded(hit)) {
			/*
			 * the cosine of the angle between light direction and surface
			 * normal
			 */
			double cosLN = -hit.normal.dotProduct(lightDir);
			/*
			 * we don't want light if the light source lies behind the object,
			 * therefore we cut off engative cosines
			 */
			diffuse = Math.max(0, cosLN);

			// perform specular lighting
			if (hit.shininess != 0 && cosLN > 0) {
				// The direction of the light ray reflected from the surface
				Vector reflectedDir = lightDir.add(hit.normal.scale(-2
						* lightDir.dotProduct(hit.normal)));
				/*
				 * the cosine of the angle between view direction and reflected
				 * ray
				 */
				double cosLR = -reflectedDir.dotProduct(r.dir);
				if (cosLR > 0)
					specular = Math.pow(cosLR, hit.shininess)
							* (hit.shininess + 2) / (2 * Math.PI);
			}
		}

		Color diffuseLight = hit.color.scale(0.7 * diffuse);
		Color specularLight = WHITE.scale(0.15 * specular);
		return ambientLight.add(diffuseLight).add(specularLight);
	}

	/*
	 * Follows the given ray until an object (sphere or wall) is hit, then
	 * performs texturing, lighting and reflection and returns the colour.
	 */
	private static Color traceRay(Ray r, int recursions) {
		PriorityQueue<Hit> hits = new PriorityQueue<Hit>();
		checkForSphereHit(r, hits);
		checkForWallHit(r, NORMAL_LEFT, 4, BLUE, null, hits);
		checkForWallHit(r, NORMAL_RIGHT, 4, BLUE, null, hits);
		checkForWallHit(r, NORMAL_UP, 2, BROWN, PARQUET_TEXTURE, hits);
		checkForWallHit(r, NORMAL_DOWN, 2, WHITE, CEILING_TEXTURE, hits);
		checkForWallHit(r, NORMAL_FRONT, 3, BLUE, null, hits);
		checkForWallHit(r, NORMAL_BACK, 5, BLUE, null, hits);

		if (hits.isEmpty()) {
			return BLACK; // background
		} else {
			// get closest hit
			Hit hit = hits.peek();

			if (hit.texture != null)
				applyTexture(hit);

			Color c = doLighting(hit, r);

			// launch reflected rays if object reflects
			if (recursions > 0 && hit.reflectivity != 0) {
				Ray rNew = new Ray();
				rNew.pos = hit.pos;
				double cosVN = r.dir.dotProduct(hit.normal);
				rNew.dir = r.dir.add(hit.normal.scale(-2 * cosVN));
				/*
				 * give new ray a head start or the rounding errors produce ugly
				 * effects as the ray hits the surface again without actually
				 * leaving it.
				 */
				rNew.pos = rNew.pos.add(rNew.dir.scale(0.001));
				c = c.scale(1 - hit.reflectivity).add(
						traceRay(rNew, recursions - 1).scale(hit.reflectivity));
			}

			return c;
		}
	}

	/*
	 * Applies the texture to the hit. Note that this is hacky and only works
	 * for the floor and the ceiling as the conversion of world coordinates to
	 * texture coordinates is hard-coded.
	 */
	private static void applyTexture(Hit hit) {
		// calculate texture coordinates from 0 to 1
		double u = (hit.pos.x + 4.0) / 8.0, v = (hit.pos.z + 5.0) / 8.0;
		if (u < 0 || u > 1 || v < 0 || v > 1)
			return;

		// calculate the texel to output
		double x = (u * (hit.texture.length - 2));
		double y = (v * (hit.texture[0].length - 2));

		/*
		 * calculate upper-left of the four texels adjacent to the position we
		 * want.
		 */
		int ix = (int) x, iy = (int) y;

		/*
		 * calculate fractional part of coordinate, i.e. fx=0.75 when our
		 * position is 3/4 into the right texel.
		 */
		double fx = x - ix, fy = y - iy;

		/*
		 * perform bilinear interpolation. we could also just return
		 * hit.texture[ix][iy], but that looks blocky up close.
		 */
		hit.color = hit.texture[ix][iy].scale((1 - fx) * (1 - fy));
		hit.color = hit.color.add(hit.texture[ix + 1][iy].scale(fx * (1 - fy)));
		hit.color = hit.color.add(hit.texture[ix][iy + 1].scale((1 - fx) * fy));
		hit.color = hit.color.add(hit.texture[ix + 1][iy + 1].scale(fx * fy));
	}

	// alternative version without bilinear interpolation
	// private static void applyTexture(Hit hit) {
	// double u = (hit.pos.x + 4.0) / 8.0, v = (hit.pos.z + 5.0) / 8.0;
	// if (u < 0 || u > 1 || v < 0 || v > 1)
	// return;
	// int x = (int) Math.round(u * (hit.texture.length - 1)), y = (int) Math
	// .round(v * (hit.texture[0].length - 1));
	// hit.color = hit.texture[x][y];
	// }

	/*
	 * Renders an image of the given dimensions with viewer at the given
	 * position and looking in the given direction. A red, reflecting sphere is
	 * placed at spherePos and walls are around it.
	 */
	public static void render(final Vector pos, final Vector view, final Vector up,
			final Vector spherePos, final BufferedImage image) {
		final int width = image.getWidth(), height = image.getHeight();
		Renderer.spherePos = spherePos;

		final int[] buffer = new int[width * height];
		// calculate right and up direction
		final Vector fRight = view.crossProduct(up).normalize();
		// ensure view/right/up are pairwise orthogonal
		final Vector fUp = fRight.crossProduct(view);

		// calculate dimensions of view plane
		final double viewPlaneHeight = 1.5;
		final double viewPlaneWidth = 1.5 * width / height;
		
		int nThreads = Runtime.getRuntime().availableProcessors();
		final int linesPerThread = height / nThreads; // number of lines handled by one thread
		Thread[] threads = new Thread[nThreads];

		// Launch a number of rendering threads
		for (int i = 0; i < nThreads; i++) {
			
			final int firstLine = i * linesPerThread; // first line to be sampled by this thread
			
			threads[i] = new Thread() { // create the i-th rendering thread
				
				public void run() {
					int k = firstLine * width; // index of first pixel to be sampled
					// sample assigned area of the view plane
					
					for (int y = firstLine; y < firstLine + linesPerThread; y++) {
						for (int x = 0; x < width; x++) {
							
							// calculate position of pixel on view plane
							double leftRight = viewPlaneWidth * (x / (double) width - 0.5);
							double upDown = viewPlaneHeight * (0.5 - y / (double) height);
							
							// calculate view ray associated with the current pixel
							Ray ray = new Ray();
							ray.pos = pos;
							ray.dir = view.add(fRight.scale(leftRight))
									.add(fUp.scale(upDown)).normalize();
							
							// perform actual ray tracing
							Color c = traceRay(ray, 1);
							buffer[k++] = c.getRGB();
						}
					}
				}
				
			};
			
			threads[i].start(); // launch created thread
		}
		
		// wait for all threads to finish
		for (Thread t : threads)
			try {
				t.join();
			} catch (InterruptedException e) {
			}

		// save result to the image
		image.setRGB(0, 0, width, height, buffer, 0, width);
	}

}
